@@ -2,16 +2,13 @@
 -- Library methods for handling MongoDB, creating and parsing packets.
 --
 -- @author Martin Holst Swende
--- @copyright Same as Nmap--See http://nmap.org/book/man-legal.html
---
--- Version 0.2
+-- @copyright Same as Nmap--See https://nmap.org/book/man-legal.html
 --
 -- @args mongodb.db - the database to use for authentication
 
 -- Created 01/13/2010 - v0.1 - created by Martin Holst Swende <martin@swende.se>
 -- Revised 01/03/2012 - v0.2 - added authentication support <patrik@cqure.net>
 
-local bin = require "bin"
 local nmap = require "nmap"
 local stdnse = require "stdnse"
 local string = require "string"
@@ -28,11 +25,11 @@ local arg_DB = stdnse.get_script_args("mongodb.db")
 -- Some lazy shortcuts
 
 local function dbg(str,...)
-  stdnse.print_debug(3, "MngoDb:"..str, ...)
+  stdnse.debug3("MngoDb:"..str, ...)
 end
---local dbg =stdnse.print_debug
+--local dbg =stdnse.debug1
 
-local err =stdnse.print_debug
+local err =stdnse.debug1
 
 ----------------------------------------------------------------------
 -- First of all comes a Bson parsing library. This can easily be moved out into a separate library should other
@@ -41,31 +38,23 @@ local err =stdnse.print_debug
 -- Library methods for handling the BSON format
 --
 -- For more documentation about the BSON format,
----and more details about it's implementations, check out the
+---and more details about its implementations, check out the
 -- python BSON implementation which is available at
 -- http://github.com/mongodb/mongo-python-driver/blob/master/pymongo/bson.py
 -- and licensed under the Apache License, Version 2.0 http://www.apache.org/licenses/LICENSE-2.0)
 --
 -- @author Martin Holst Swende
--- @copyright Same as Nmap--See http://nmap.org/book/man-legal.html
+-- @copyright Same as Nmap--See https://nmap.org/book/man-legal.html
 --
 -- Version 0.1
 
 -- Created 01/13/2010 - v0.1 - created by Martin Holst Swende <martin@swende.se>
 --module("bson", package.seeall)
---require("bin")
 local function dbg_err(str,...)
-  stdnse.print_debug("Bson-ERR:"..str, ...)
+  stdnse.debug1("Bson-ERR:"..str, ...)
 end
 --local err =stdnse.log_error
 
--- Packs data into nullterminated string
---@param input the string to pack
---@return the packed nullterminated string
-local function make_nullterminated_string(input)
-  return bin.pack("z",input)
-end
-
 --Converts an element (key, value) into bson binary data
 --@param key the key name, must *NOT* contain . (period) or start with $
 --@param value, the element value
@@ -84,20 +73,18 @@ local function _element_to_bson(key, value)
     return false, ("key %r must not contain '.'"):format(tostring(key))
   end
 
-  local name =bin.pack("z",key) -- null-terminated string
+  local name = string.pack("z", key) -- null-terminated string
   if type(value) == 'string' then
-    local cstring = bin.pack("z",value) -- null-terminated string
-    local length = bin.pack("<i", cstring:len())
-    local op = bin.pack('H','02')
+    local cstring = string.pack("z", value) -- null-terminated string
+    local length = string.pack("<i4", cstring:len())
+    local op = "\x02"
     return true, op .. name .. length .. cstring
   elseif type(value) =='table' then
-    return true, bin.pack('H','02') .. name .. toBson(value)
+    return true, "\x02" .. name .. toBson(value)
   elseif type(value)== 'boolean' then
-    return true, bin.pack('H','08') + name + bin.pack('H',value and '01' or '00')
+    return true, "\x08" .. name .. (value and '\x01' or '\0')
   elseif type(value) == 'number' then
-    --return bin.pack('H','10').. name .. bin.pack("<i", value)
-    -- Use 01 - double for - works better than 10
-    return true, bin.pack('H','01') .. name .. bin.pack("<d", value)
+    return true, '\x01' .. name .. string.pack("<d", value)
   end
 
   local _ = ("cannot convert value of type %s to bson"):format(type(value))
@@ -140,7 +127,7 @@ function toBson(dict)
   end
   dbg("Packet length is %d",length)
   --Final pack
-  return true, bin.pack("I", length) .. elements .. bin.pack('H',"00")
+  return true, string.pack("<I4", length) .. elements .. "\0"
 end
 
 -- Reads a null-terminated string. If length is supplied, it is just cut
@@ -151,7 +138,7 @@ end
 --@return the remaining data (*without* null-char)
 local function get_c_string(data,length)
   if not length then
-    local index = data:find(string.char(0))
+    local index = data:find('\0')
     if index == nil then
       error({code="C-string did not contain NULL char"})
     end
@@ -171,10 +158,11 @@ end
 -- @return error string if error occurred
 local function parse(code,data)
   if 1 == code  then -- double
-    return bin.unpack("<d", data)
+    local v, pos = string.unpack("<d", data)
+    return pos, v
   elseif 2 == code then -- string
     -- data length = first four bytes
-    local _,len = bin.unpack("<i",data)
+    local len = string.unpack("<i4",data)
     -- string data = data[5] -->
     local value = get_c_string(data:sub(5), len)
     -- Count position as header (=4) + length of string (=len)+ null char (=1)
@@ -183,7 +171,7 @@ local function parse(code,data)
     local object, err
 
     -- Need to know the length, to return later
-    local _,obj_size = bin.unpack("<i", data)
+    local obj_size = string.unpack("<i4", data)
     -- Now, get the data object
     dbg("Recursing into bson array")
     object, data, err = fromBson(data)
@@ -195,7 +183,8 @@ local function parse(code,data)
   elseif 8 == code then -- Boolean
     return 2, data:byte(1) == 1
   elseif 9 == code then -- int64, UTC datetime
-    return bin.unpack("<l", data)
+    local v, pos = string.unpack("<i8", data)
+    return pos, v
   elseif 10 == code then -- nullvalue
     return 0,nil
     --11= _get_regex
@@ -204,10 +193,12 @@ local function parse(code,data)
     --14= _get_string, # symbol
     --15=  _get_code_w_scope
   elseif 16 == code then -- 4 byte integer
-    return bin.unpack("<i", data)
+    local v, pos = string.unpack("<i4", data)
+    return pos, v
     --17= _get_timestamp
   elseif 18 == code then -- long
-    return bin.unpack("<l", data)
+    local v, pos = string.unpack("<i8", data)
+    return pos, v
   end
   local err = ("Getter for %d not implemented"):format(code)
   return 0, data, err
@@ -266,7 +257,7 @@ function isPacketComplete(data)
     return false
   end
 
-  local _,obj_size = bin.unpack("<i", data)
+  local obj_size = string.unpack("<i4", data)
 
   dbg("BSon packet size is %s", obj_size)
 
@@ -382,15 +373,15 @@ MongoData ={
     return o
   end
 }
---Adds unsigned int32 to the message body
+--Adds signed int32 to the message body
 --@param value the value to add
-function MongoData:addUnsignedInt32(value)
-  self.valueString = self.valueString..bin.pack("I",value)
+function MongoData:addInt32(value)
+  self.valueString = self.valueString..string.pack("<i4",value)
 end
 -- Adds a string to the message body
 --@param value the string to add
 function MongoData:addString(value)
-  self.valueString = self.valueString..bin.pack('z',value)
+  self.valueString = self.valueString..string.pack('z',value)
 end
 -- Add a table as a BSon object to the body
 --@param dict the table to be converted to BSon
@@ -411,10 +402,10 @@ end
 -- This method creates necessary header information and puts it with the body
 function MongoData:data()
   local header = MongoData:new()
-  header:addUnsignedInt32( self.valueString:len()+4+4+4+4)
-  header:addUnsignedInt32( self.requestID)
-  header:addUnsignedInt32( self.responseTo or 0xFFFFFFFF)
-  header:addUnsignedInt32( self.opCode)
+  header:addInt32( self.valueString:len()+4+4+4+4)
+  header:addInt32( self.requestID)
+  header:addInt32( self.responseTo or -1)
+  header:addInt32( self.opCode)
   return header.valueString .. self.valueString
 end
 -- Creates a query
@@ -424,10 +415,13 @@ end
 --@return packet data OR error message
 local function createQuery(collectionName, query)
   local packet = MongoData:new({opCode=MongoData.OP.QUERY})
-  packet:addUnsignedInt32(0); -- options
+  packet:addInt32(0); -- options
   packet:addString(collectionName);
-  packet:addUnsignedInt32(0) -- number to skip
-  packet:addUnsignedInt32(-1) -- number to return : no limit
+  packet:addInt32(0) -- number to skip
+  -- NB: Using value of -1 for "no limit" below is suspect. The protocol
+  --     interprets -1 as requesting only one document, not all documents.
+  --     https://docs.mongodb.com/manual/reference/mongodb-wire-protocol/#wire-op-query
+  packet:addInt32(-1) -- number to return : no limit
   local status, error = packet:addBSON(query)
 
   if not status then
@@ -487,11 +481,11 @@ end
 --@return int32 value
 --@return data unread
 local function parseInt32(data)
-  local pos,val = bin.unpack("<i",data)
+  local val, pos = string.unpack("<i4", data)
   return val, data:sub(pos)
 end
 local function parseInt64(data)
-  local pos,val =  bin.unpack("<l",data)
+  local val, pos =  string.unpack("<i8", data)
   return val, data:sub(pos)
 end
 -- Parses response header
@@ -538,7 +532,7 @@ function isPacketComplete(data)
     return false
   end
 
-  local _,obj_size = bin.unpack("<i", data)
+  local obj_size = string.unpack("<i4", data)
 
   dbg("MongoDb Packet size is %s, (got %d)", obj_size,data:len())
 
@@ -558,7 +552,7 @@ function query(socket, data)
   --Create an error handler
   local catch = function()
     socket:close()
-    stdnse.print_debug("Query failed")
+    stdnse.debug1("Query failed")
   end
   local try = nmap.new_try(catch)
 
@@ -646,42 +640,5 @@ function queryResultToTable( resultTable )
   return result
 
 end
-----------------------------------------------------------------------------------
--- Test-code for debugging purposes below
-----------------------------------------------------------------------------------
-
-
---- Prints data (string) as Hex values, e.g so it more easily can
--- be compared with a packet dump
--- @param strData the data in a string format
-
-local function printBuffer(strData)
-  local out = ''
-  local ch
-  for i  = 1,strData:len() do
-    out = out .." "
-    ch =strData:byte(i)
-    if(ch < 16) then
-      ch = string.format("0%x",ch)
-    else ch = string.format("%x",ch)
-    end
-    --if ch > 64 and ch < 123 then
-    --  out = out .. string.char(ch)
-    --else
-    out = out .. ch
-    --end
-  end
-  print(out)
-end
-
--- function test()
---   local res
---   res = versionQuery()
---   print(type(res),res:len(),res)
---   local  out= bin.unpack('C'..#res,res)
---   printBuffer(res)
--- end
---test()
-
 
 return _ENV;
