@@ -12,6 +12,7 @@
 -- * IMAP
 -- * LDAP
 -- * NNTP
+-- * MySQL
 -- * POP3
 -- * PostgreSQL
 -- * SMTP
@@ -22,30 +23,47 @@
 -- @author Patrik Karlsson <patrik@cqure.net>
 
 local asn1 = require "asn1"
-local bin = require "bin"
 local comm = require "comm"
 local ftp = require "ftp"
 local ldap = require "ldap"
 local match = require "match"
 local mssql = require "mssql"
+local mysql = require "mysql"
 local nmap = require "nmap"
 local smtp = require "smtp"
 local stdnse = require "stdnse"
 local string = require "string"
 local table = require "table"
+local tableaux = require "tableaux"
 local tls = require "tls"
 local vnc = require "vnc"
 local xmpp = require "xmpp"
+local have_openssl, openssl = pcall(require, "openssl")
 _ENV = stdnse.module("sslcert", stdnse.seeall)
 
---- Parse an X.509 certificate from DER-encoded string
---@name parse_ssl_certificate
---@class function
---@param der DER-encoded certificate
---@return table containing decoded certificate or nil on failure
---@return error string if parsing failed
---@see nmap.get_ssl_certificate
-_ENV.parse_ssl_certificate = nmap.socket.parse_ssl_certificate
+if have_openssl then
+  --- Parse an X.509 certificate from DER-encoded string
+  --
+  -- This uses OpenSSL's X.509 parsing routines, so if OpenSSL support is not
+  -- included, only the <code>pem</code> key of the returned table will be
+  -- present.
+  --@name parse_ssl_certificate
+  --@class function
+  --@param der DER-encoded certificate
+  --@return table containing decoded certificate or nil on failure
+  --@return error string if parsing failed
+  --@see nmap.get_ssl_certificate
+  _ENV.parse_ssl_certificate = nmap.socket.parse_ssl_certificate
+else
+  local base64 = require "base64"
+  _ENV.parse_ssl_certificate = function(der)
+    return {
+      pem = ("-----BEGIN CERTIFICATE-----\n%s\n-----END CERTIFICATE-----\n"):format(
+        base64.enc(der):gsub("(" .. ("."):rep(64) .. ")", "%1\n"):gsub("\n$", "")
+        )
+    }
+  end
+end
 
 -- Mark whether this port supports STARTTLS, to save connection attempts later.
 -- If it ever succeeds, it can't be marked as failing later, but if it fails
@@ -283,7 +301,7 @@ StartTLS = {
 
     -- 0x80  = 10000001  =  10        0                 00000
     -- hex     binary       Context   Primitive value   Field: requestName  Value: 0
-    local encodedOID = bin.pack('HAA' , '80', string.char(#oid), oid)
+    local encodedOID = string.pack('Bs1', 0x80, oid)
 
     local ldapRequest, ldapRequestId
     local ExtendedRequest = 23
@@ -346,7 +364,7 @@ StartTLS = {
     end
 
     local status
-    status, result = smtp.query(s, "LHLO", "example.com")
+    status, result = smtp.query(s, "LHLO", smtp.get_domain(host))
     if not status then
       stdnse.debug1("LHLO with errors or timeout.  Enable --script-trace to see what is happening.")
       return false, string.format("Failed to LHLO: %s", result)
@@ -378,6 +396,40 @@ StartTLS = {
 
   lmtp_prepare_tls = tls_reconnect("lmtp_prepare_tls_without_reconnect"),
 
+  mysql_prepare_tls_without_reconnect = function(host, port)
+    local s, err = comm.opencon(host, port)
+    if not s then
+      return false, string.format("Failed to connect to MySQL server: %s", err)
+    end
+    local status, resp = mysql.receiveGreeting(s)
+    if not status then
+      return false, string.format("MySQL handshake error: %s", resp)
+    end
+    if 0 == resp.capabilities & mysql.Capabilities.SwitchToSSLAfterHandshake then
+      return false, "MySQL server does not support SSL"
+    end
+    local clicap = mysql.Capabilities.SwitchToSSLAfterHandshake
+    + mysql.Capabilities.LongPassword
+    + mysql.Capabilities.LongColumnFlag
+    + mysql.Capabilities.SupportsLoadDataLocal
+    + mysql.Capabilities.Speaks41ProtocolNew
+    + mysql.Capabilities.InteractiveClient
+    + mysql.Capabilities.SupportsTransactions
+    + mysql.Capabilities.Support41Auth
+    local packet = string.pack( "<I2I2I4B c23",
+      clicap,
+      0,
+      16777216,
+      mysql.Charset.latin1_COLLATE_latin1_swedish_ci,
+      string.rep("\0", 23)
+      )
+    packet = string.pack("<I4", #packet + (1 << 24)) .. packet
+    s:send(packet)
+    return true, s
+  end,
+
+  mysql_prepare_tls = tls_reconnect("mysql_prepare_tls_without_reconnect"),
+
   nntp_prepare_tls_without_reconnect = function(host, port)
     local s, err, result = comm.opencon(host, port, "", {lines=1, recv_before=true})
     if not s then
@@ -440,7 +492,7 @@ StartTLS = {
   postgres_prepare_tls_without_reconnect = function(host, port)
     -- http://www.postgresql.org/docs/devel/static/protocol-message-formats.html
     -- 80877103 is "SSLRequest" in v2 and v3 of Postgres protocol
-    local s, resp = comm.opencon(host, port, bin.pack(">II", 8, 80877103))
+    local s, resp = comm.opencon(host, port, string.pack(">I4I4", 8, 80877103))
     if not s then
       return false, ("Failed to connect to Postgres server: %s"):format(resp)
     end
@@ -468,7 +520,7 @@ StartTLS = {
     end
 
     local status
-    status, result = smtp.ehlo(s, "example.com")
+    status, result = smtp.ehlo(s, smtp.get_domain(host))
     if not status then
       stdnse.debug1("EHLO with errors or timeout.  Enable --script-trace to see what is happening.")
       return false, string.format("Failed to connect to SMTP server: %s", result)
@@ -509,14 +561,14 @@ StartTLS = {
     if not status then return status, preloginResponse end
 
     local encryption
-    local pos, optype, oppos, oplen = bin.unpack('>CSS', result)
+    local optype, oppos, oplen, pos = string.unpack('>BI2I2', result)
     while optype ~= mssql.PreLoginPacket.OPTION_TYPE.Terminator do
       --stdnse.debug1("optype: %d, oppos: %x, oplen: %d", optype, oppos, oplen)
       if optype == mssql.PreLoginPacket.OPTION_TYPE.Encryption then
-        pos, encryption = bin.unpack('C', result, oppos + 1)
+        encryption, pos = string.unpack('B', result, oppos + 1)
         break
       end
-      pos, optype, oppos, oplen = bin.unpack('>CSS', result, pos)
+      optype, oppos, oplen, pos = string.unpack('>BI2I2', result, pos)
     end
     if not encryption then
       starttls_supported(host, port, false)
@@ -564,9 +616,9 @@ StartTLS = {
 
             -- read in the TDS headers
             local packetType, messageStatus, packetLength
-            pos, packetType, messageStatus, packetLength = bin.unpack(">CCS", readBuffer, pos )
+            packetType, messageStatus, packetLength, pos = string.unpack(">BBI2", readBuffer, pos )
             local spid, packetId, window
-            pos, spid, packetId, window = bin.unpack(">SCC", readBuffer, pos )
+            spid, packetId, window, pos = string.unpack(">I2BB", readBuffer, pos )
 
             if packetLength > #readBuffer then
               status, result = tds._socket:receive_bytes(packetLength - #readBuffer)
@@ -638,7 +690,7 @@ StartTLS = {
       }
       local best
       for i=1, #auth_order do
-        if stdnse.contains(v.vencrypt.types, auth_order[i]) then
+        if tableaux.contains(v.vencrypt.types, auth_order[i]) then
           best = auth_order[i]
           break
         end
@@ -648,7 +700,7 @@ StartTLS = {
         starttls_supported(host, port, false)
         return false, "No TLS VeNCrypt auth subtype received"
       end
-      sock:send(bin.pack(">I", best))
+      sock:send(string.pack(">I4", best))
       local status, buf = sock:receive_buf(match.numbytes(1), true)
       if not status or string.byte(buf, 1) ~= 1 then
         starttls_supported(host, port, false)
@@ -657,7 +709,7 @@ StartTLS = {
       starttls_supported(host, port, true)
       return true, sock
     elseif v:supportsSecType(vnc.VNC.sectypes.TLS) then
-      status = sock:send( bin.pack("C", vnc.VNC.sectypes.TLS) )
+      status = sock:send( string.pack("B", vnc.VNC.sectypes.TLS) )
       if not status then
         starttls_supported(host, port, false)
         return false, "Failed to select TLS authentication type"
@@ -766,6 +818,8 @@ local SPECIALIZED_PREPARE_TLS = {
   smtp = StartTLS.smtp_prepare_tls,
   [25] = StartTLS.smtp_prepare_tls,
   [587] = StartTLS.smtp_prepare_tls,
+  mysql = StartTLS.mysql_prepare_tls,
+  [3306] = StartTLS.mysql_prepare_tls,
   xmpp = StartTLS.xmpp_prepare_tls,
   [5222] = StartTLS.xmpp_prepare_tls,
   [5269] = StartTLS.xmpp_prepare_tls,
@@ -791,6 +845,8 @@ local SPECIALIZED_PREPARE_TLS_WITHOUT_RECONNECT = {
   smtp = StartTLS.smtp_prepare_tls_without_reconnect,
   [25] = StartTLS.smtp_prepare_tls_without_reconnect,
   [587] = StartTLS.smtp_prepare_tls_without_reconnect,
+  mysql = StartTLS.mysql_prepare_tls_without_reconnect,
+  [3306] = StartTLS.mysql_prepare_tls_without_reconnect,
   xmpp = StartTLS.xmpp_prepare_tls_without_reconnect,
   [5222] = StartTLS.xmpp_prepare_tls_without_reconnect,
   [5269] = StartTLS.xmpp_prepare_tls_without_reconnect,
@@ -863,9 +919,12 @@ local function get_record_iter(sock)
   local i = 1
   local fragment
   return function ()
-    local record
-    i, record = tls.record_read(buffer, i, fragment)
+    local record, more
+    i, record, more = tls.record_read(buffer, i, fragment)
     if record == nil then
+      if not more then
+        return nil, "no more"
+      end
       local status, err
       status, buffer, err = tls.record_buffer(sock, buffer, i)
       if not status then
@@ -881,6 +940,57 @@ local function get_record_iter(sock)
   end
 end
 
+local function handshake_cert (socket)
+    -- logic mostly lifted from ssl-enum-ciphers
+    local hello = tls.client_hello()
+    local status, err = socket:send(hello)
+    if not status then
+      return false, "Failed to send to server"
+    end
+
+    local get_next_record = get_record_iter(socket)
+    local records = {}
+    local done = false
+    while not done do
+      local record
+      record, err = get_next_record()
+      if not record then
+        stdnse.debug1("no record: %s", err)
+        break
+      end
+      -- Collect message bodies into one record per type
+      records[record.type] = records[record.type] or record
+      for j = 1, #record.body do -- no ipairs because we append below
+        local b = record.body[j]
+        done = ((record.type == "alert" and b.level == "fatal") or
+          (record.type == "handshake" and b.type == "server_hello_done"))
+        table.insert(records[record.type].body, b)
+      end
+    end
+
+    local handshake = records.handshake
+    if not handshake then
+      return false, "Server did not handshake"
+    end
+
+    local certs
+    for i, b in ipairs(handshake.body) do
+      if b.type == "certificate" then
+        certs = b
+        break
+      end
+    end
+    if not certs or not next(certs.certificates) then
+      return false, "Server sent no certificate"
+    end
+
+    local cert, err = parse_ssl_certificate(certs.certificates[1])
+    if not cert then
+      return false, ("Unable to parse cert: %s"):format(err)
+    end
+    return true, cert
+end
+
 --- Gets a certificate for the given host and port
 -- The function will attempt to START-TLS for the ports known to require it.
 -- @param host table as received by the script action function
@@ -901,106 +1011,75 @@ function getCertificate(host, port)
 
   local cert
 
+  local wrapper = SPECIALIZED_WRAPPED_TLS_WITHOUT_RECONNECT[port.service] or SPECIALIZED_WRAPPED_TLS_WITHOUT_RECONNECT[port.number]
+  local special_table = have_openssl and SPECIALIZED_PREPARE_TLS or SPECIALIZED_PREPARE_TLS_WITHOUT_RECONNECT
+  local specialized = special_table[port.service] or special_table[port.number]
+
+  local status = false
+
   -- If we don't already know the service is TLS wrapped check to see if we
   -- have to use a wrapper and do a manual handshake
-  local wrapper
-  if not ( port.version.service_tunnel == 'ssl' ) then
-    wrapper = SPECIALIZED_WRAPPED_TLS_WITHOUT_RECONNECT[port.service] or SPECIALIZED_WRAPPED_TLS_WITHOUT_RECONNECT[port.number]
+  if wrapper and port.version.service_tunnel ~= 'ssl' then
+    local socket
+    status, socket = wrapper(host, port)
+    if not status then
+      stdnse.debug1("Wrapper function error: %s", socket)
+    else
+      status, cert = handshake_cert(socket)
+      socket:close()
+    end
   end
 
-  if wrapper then
-    local status, socket = wrapper(host, port)
+  -- If that didn't work, see if we need a specialized connection method
+  if not status and specialized and port.version.service_tunnel ~= 'ssl' then
+    local socket
+    status, socket = specialized(host, port)
     if not status then
-      mutex "done"
-      return false, socket
-    end
-
-    -- logic mostly lifted from ssl-enum-ciphers
-    local hello = tls.client_hello()
-    local status, err = socket:send(hello)
-    if not status then
-      mutex "done"
-      return false, "Failed to connect to server"
-    end
-
-    local get_next_record = get_record_iter(socket)
-    local records = {}
-    while true do
-      local record
-      record, err = get_next_record()
-      if not record then
-        stdnse.debug1("no record: %s", err)
-        socket:close()
-        break
-      end
-      -- Collect message bodies into one record per type
-      records[record.type] = records[record.type] or record
-      local done = false
-      for j = 1, #record.body do -- no ipairs because we append below
-        local b = record.body[j]
-        done = ((record.type == "alert" and b.level == "fatal") or
-          (record.type == "handshake" and b.type == "server_hello_done"))
-        table.insert(records[record.type].body, b)
-      end
-      if done then
-        socket:close()
-        break
-      end
-    end
-
-    local handshake = records.handshake
-    if not handshake then
-      mutex "done"
-      return false, "Server did not handshake"
-    end
-
-    local certs
-    for i, b in ipairs(handshake.body) do
-      if b.type == "certificate" then
-        certs = b
-        break
-      end
-    end
-    if not certs or not next(certs.certificates) then
-      mutex "done"
-      return false, "Server sent no certificate"
-    end
-
-    cert, err = parse_ssl_certificate(certs.certificates[1])
-    if not cert then
-      mutex "done"
-      return false, ("Unable to get cert: %s"):format(err)
-    end
-  else
-    -- If we don't already know the service is TLS wrapped check to see if
-    -- there a specialized function for this port
-    local specialized
-    if not ( port.version.service_tunnel == 'ssl' ) then
-      specialized = SPECIALIZED_PREPARE_TLS[port.service] or SPECIALIZED_PREPARE_TLS[port.number]
-    end
-    local status
-    local socket = nmap.new_socket()
-    if specialized then
-      status, socket = specialized(host, port)
-      if not status then
-        mutex "done"
-        stdnse.debug1("Specialized function error: %s", socket)
-        return false, "Failed to connect to server"
-      end
+      stdnse.debug1("Specialized function error: %s", socket)
     else
-      status = socket:connect(host, port, "ssl")
-      if ( not(status) ) then
-        mutex "done"
-        return false, "Failed to connect to server"
+      if have_openssl then
+        cert = socket:get_ssl_certificate()
+        status = not not cert
+      else
+        status, cert = handshake_cert(socket)
       end
+      socket:close()
     end
-    cert = socket:get_ssl_certificate()
-    if ( cert == nil ) then
-      mutex "done"
-      return false, "Unable to get cert"
+  end
+
+  -- Now try to connect with Nsock's SSL connection
+  if not status and have_openssl then
+    local socket = nmap.new_socket()
+    local errmsg
+    status, errmsg = socket:connect(host, port, "ssl")
+    if not status then
+      stdnse.debug1("SSL connect error: %s", errmsg)
+    else
+      cert = socket:get_ssl_certificate()
+      status = not not cert
+      socket:close()
     end
   end
 
+  -- Finally, try to connect and manually handshake (maybe more tolerant of TLS
+  -- insecurity than OpenSSL)
+  if not status then
+    local socket = nmap.new_socket()
+    local errmsg
+    status, errmsg = socket:connect(host, port)
+    if not status then
+      stdnse.debug1("Connect error: %s", errmsg)
+    else
+      status, cert = handshake_cert(socket)
+      socket:close()
+    end
+  end
+
+  if not status then
+    mutex "done"
+    return false, "No certificate found"
+  end
+
   host.registry["ssl-cert"] = host.registry["ssl-cert"] or {}
   host.registry["ssl-cert"][port.number] = host.registry["ssl-cert"][port.number] or {}
   host.registry["ssl-cert"][port.number] = cert
